/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "AVLTree.h"
#include "AVLTree_priv.h"

// A payload delete function that does nothing; this turns out
// to be useful in a subcase of RecursiveRemoveFromAVLTree().
void NullAVLPayloadDelete(void *pl) { }

// A helper function for AVLTreeDelete; is O(N)
void RecursiveAVLTreeDelete(AVLTreeNodePtr tn,
                            TreePayloadFreeFnPtr pffp);

// A helper function for InsertIntoAVLTree; is O(log(N)).  Does
// an insertion, but doesn't try to preserve balance; instead,
// it always inserts as a leaf node.  The caller must invoke
// Rebalance() after using this.
AVLTreeNodePtr RecursiveInsertIntoAVLTree(AVLTreeNodePtr tn,
                                          void *pl,
                                          TreePayloadComparatorFnPtr cfp,
                                          AVLTreePtr tree);

// A helper function for LookupInAVLTree; is O(logN).
int RecursiveLookupInAVLTree(AVLTreeNodePtr tn,
                             void *payload_search,
                             void **payload_return,
                             TreePayloadComparatorFnPtr cfp);

// A helper function for RemoveFromAVLTree; is O(log(N)).  Does
// a removal and also preserves balance.
int RecursiveRemoveFromAVLTree(AVLTreeNodePtr tn,
                               void *pl,
                               TreePayloadComparatorFnPtr cfp,
                               TreePayloadFreeFnPtr pffp,
                               AVLTreePtr tree);

// Recalculate the heights of tree nodes, starting at tn and working
// up to the tree's root; is O(log(N)).
void ResetHeightUpwards(AVLTreeNodePtr tn);

// Rebalance a tree, starting at tn and working up to root, performing
// rotations as needed along the way.  Is O(log(N)).
void Rebalance(AVLTreeNodePtr tn, AVLTreePtr tree);

// The four rotation cases.
void RR(AVLTreeNodePtr tn_3);
void RL(AVLTreeNodePtr tn_3);
void LR(AVLTreeNodePtr tn_5);
void LL(AVLTreeNodePtr tn_5);

// Get the successor of a node, or NULL if there is none.
AVLTreeNodePtr GetSuccessor(AVLTreeNodePtr tn);

// Get the predecessor of a node, or NULL if there is none.
AVLTreeNodePtr GetPredecessor(AVLTreeNodePtr tn);

// Allocate and initialize a new AVLTreeNodePtr, setting
// its parent to 'parent' and payload to 'pl'.
AVLTreeNodePtr AllocAVLTreeNode(AVLTreeNodePtr parent, void *pl);

// Utility function to absorb the common logic in
// AVLTreeIteratorHasNext and AVLTreeIteratorNext.
int AVLTreeIteratorDoNext(AVLTreeIterPtr iter, int do_advance);

// Utility function to delete a leaf or single-child node.
void RemoveLeafOrChildFromAVLTree(AVLTreeNodePtr tn,
                                  TreePayloadFreeFnPtr pffp,
                                  AVLTreePtr tree);

//
// Public functions from AVLTree.h.
//

AVLTreePtr AllocateAVLTree(void) {
  // allocate the tree record
  AVLTreePtr t = (AVLTreePtr) malloc(sizeof(AVLTree));
  if (t == NULL) {
    // out of memory
    return (AVLTreePtr) NULL;
  }

  // initialize the newly allocated record structure
  t->num_nodes = 0;
  t->root = NULL;

  // return the newly minted and empty tree
  return t;
}

void FreeAVLTree(AVLTreePtr tree,
                 TreePayloadFreeFnPtr payload_free_function) {
  // defensive programming
  assert(tree != NULL);
  assert(payload_free_function != NULL);

  // recursive delete
  RecursiveAVLTreeDelete(tree->root, payload_free_function);

  // delete the record
  free(tree);
}

uint64_t NumElementsInAVLTree(AVLTreePtr tree) {
  // defensive programming
  assert(tree != NULL);
  return tree->num_nodes;
}

int InsertIntoAVLTree(AVLTreePtr tree,
                      void *payload,
                      TreePayloadComparatorFnPtr comparator_function) {
  // defensive programming
  assert(tree != NULL);
  assert(comparator_function != NULL);

  // if this is an empty tree, do the insert at the root
  if (tree->root == NULL) {
    assert(tree->num_nodes == 0);
    tree->root = AllocAVLTreeNode(NULL, payload);
    tree->num_nodes = 1;
    return 1;
  }

  // the tree is not null, so do the recursive descent and insertion,
  // and receive a pointer to the inserted node back
  AVLTreeNodePtr np = RecursiveInsertIntoAVLTree(tree->root,
                                                 payload,
                                                 comparator_function,
                                                 tree);
  if (np == NULL) {
    // nothing was inserted; malloc must have failed or the
    // payload already exists in the tree
    return 0;
  }
  // a new node was inserted, so increment the tree node count
  tree->num_nodes += 1;

  // the insert worked, but we might have become unbalanced.  so, we
  // need to check the balance starting from the newly inserted node,
  // back up to the root.
  Rebalance(np, tree);

  // success!
  return 1;
}

int LookupInAVLTree(AVLTreePtr tree,
                    void *payload_search,
                    void **payload_return,
                    TreePayloadComparatorFnPtr comparator_function) {
  // defensive programming
  assert(tree != NULL);
  assert(comparator_function != NULL);

  // do an inorder traversal
  return RecursiveLookupInAVLTree(tree->root,
                                  payload_search,
                                  payload_return,
                                  comparator_function);
}

int RemoveFromAVLTree(AVLTreePtr tree,
                      void *remove_me,
                      TreePayloadComparatorFnPtr comparator_function,
                      TreePayloadFreeFnPtr payload_free_function) {
  // defensive programming
  assert(tree != NULL);
  assert(comparator_function != NULL);
  assert(payload_free_function != NULL);

  // do the recursive descent and delete, and receive and return
  // success
  int res = RecursiveRemoveFromAVLTree(tree->root,
                                       remove_me,
                                       comparator_function,
                                       payload_free_function,
                                       tree);
  // if success, decrement the tree node count
  if (res)
    tree->num_nodes -= 1;
  return res;
}

AVLTreeIterPtr AVLTreeMakeIterator(AVLTreePtr tree,
                                   AVLTreeIteratorOrder order) {
  // defensive programming
  assert(tree != NULL);
  assert((order >= INORDER) || (order <= PREORDER));

  // if the tree is empty, return a null iterator
  if (NumElementsInAVLTree(tree) == 0)
    return NULL;

  // manufacture an iterator
  AVLTreeIterPtr ti = (AVLTreeIterPtr) malloc(sizeof(AVLTreeIter));
  if (ti == NULL) {
    // out of memory!
    return NULL;
  }
  // set up the iterator
  ti->tree = tree;
  ti->node = tree->root;
  ti->order = order;

  switch (ti->order) {
  case INORDER:
    // start the traversal at the smallest node (left-most descendant)
    while (ti->node->left != NULL)
      ti->node = ti->node->left;
    break;
  case INORDER_REVERSE:
    // start the traversal at the largest node (right-most descendant)
    while (ti->node->right != NULL)
      ti->node = ti->node->right;
    break;
  case PREORDER:
    // start the traversal at the root
    break;
  default:
    // should never get here
    assert(0);
  }
  return ti;
}

void AVLTreeIteratorFree(AVLTreeIterPtr iter) {
  // defensive programming
  assert(iter != NULL);

  // do the memory deallocation
  free(iter);
}

int AVLTreeIteratorHasNext(AVLTreeIterPtr iter) {
  return AVLTreeIteratorDoNext(iter, 0);
}

int AVLTreeIteratorNext(AVLTreeIterPtr iter) {
  return AVLTreeIteratorDoNext(iter, 1);
}

void AVLTreeIteratorGetPayload(AVLTreeIterPtr iter, void **payload) {
  // defensive programming
  assert(iter != NULL);
  assert(iter->node != NULL);

  // return the payload
  *payload = iter->node->payload;
}

//
// Private functions
//

void RecursiveAVLTreeDelete(AVLTreeNodePtr t, TreePayloadFreeFnPtr pffp) {
  // Are we past a leaf?  If so, stop the recursion.
  if (t == NULL)
    return;

  // do a post-order traversal deletion
  RecursiveAVLTreeDelete(t->left, pffp);
  RecursiveAVLTreeDelete(t->right, pffp);

  // post-order free -- both the payload and the node
  pffp(t->payload);
  free(t);
}

AVLTreeNodePtr RecursiveInsertIntoAVLTree(AVLTreeNodePtr tn,
                                          void *pl,
                                          TreePayloadComparatorFnPtr cfp,
                                          AVLTreePtr tree) {
  // do a little defensive programming.
  assert(tn != NULL);
  assert(tree != NULL);

  // figure out which way to descend
  int direction = cfp(pl, tn->payload);
  if (direction == 0) {
    // this payload is already in the tree, so fail; can't insert it twice.
    return NULL;
  }

  // go left?
  if (direction == -1) {
    // Yup; is there a left tree to descend into?  If so, do it.
    if (tn->left != NULL) {
      return RecursiveInsertIntoAVLTree(tn->left, pl, cfp, tree);
    }

    // hit the bottom, so allocate and insert a new left leaf node
    tn->left = AllocAVLTreeNode(tn, pl);
    if (tn->left == NULL)
      return NULL;  // out of memory

    // We added left.  Potentially reset the height of the new node and
    // upwards through the tree.
    ResetHeightUpwards(tn->left);
    return tn->left;
  }

  // direction == +1, i.e., right case
  assert(direction == 1);
  if (tn->right != NULL) {
    return RecursiveInsertIntoAVLTree(tn->right, pl, cfp, tree);
  }

  // allocate and insert here
  tn->right = AllocAVLTreeNode(tn, pl);
  if (tn->right == NULL)
    return NULL;  // out of memory

  // We added right.  Recalculate the height of the new node and
  // upwards through the tree.
  ResetHeightUpwards(tn->right);
  return tn->right;
}

AVLTreeNodePtr AllocAVLTreeNode(AVLTreeNodePtr parent, void *pl) {
  // do the malloc.
  AVLTreeNodePtr ret = (AVLTreeNodePtr) malloc(sizeof(AVLTreeNode));
  if (ret == NULL)
    return NULL;  // out of memory

  // initialize the fields.
  ret->parent = parent;
  ret->left = NULL;
  ret->right = NULL;
  ret->payload = pl;
  ret->height = 1;
  return ret;
}

void Rebalance(AVLTreeNodePtr tn, AVLTreePtr tree) {
  // defensive programming
  assert(tn != NULL);

  // if my balance is (-1, 0, or +1), tn is OK, so move up the tree.
  if ((GetBalance(tn) == -1) ||
      (GetBalance(tn) == 0) ||
      (GetBalance(tn) == +1)) {
    if (tn->parent != NULL)
      Rebalance(tn->parent, tree);
    return;
  }

  // tn is out of balance!  need to figure out which rotation to do to
  // fix it.
  if (GetBalance(tn) == -2) {
    assert(tn->right != NULL);
    // Check RL vs. RR.
    if (GetBalance(tn->right) > 0) {
      // do RL rotation; this does not change the root
      // in any circumstance.
      RL(tn);
    }
    // Either RL put us into RR case, or we were there before.
    // Do RR rotation.
    RR(tn);

    // If we were the root, then the tree root needs
    // to be updated to my new parent.
    if (tree->root == tn) {
      tree->root = tn->parent;
    }
  } else if (GetBalance(tn) == +2) {
    assert(tn->left != NULL);
    // Check LR / LL.
    if (GetBalance(tn->left) < 0) {
      // do LR rotation; tree root will never change.
      LR(tn);
    }
    // Either LR put us into LL case, or we were there before.
    // Do LL rotation.
    LL(tn);

    // If we were the root, then the tree root needs
    // to be updated to my new parent.
    if (tree->root == tn) {
      tree->root = tn->parent;
    }
  } else {
    // Bbalance is not in range -2 .. +2, so something went
    // horribly wrong.
    assert(0);
    return;
  }

  // continue rebalancing up through parent towards root
  if (tn->parent != NULL)
    Rebalance(tn->parent, tree);
}

void RR(AVLTreeNodePtr tn_3) {
  AVLTreeNodePtr tn_4 = tn_3->right, tn_b = tn_4->left;

  // move b to be the right-child of tn.
  tn_3->right = tn_b;
  if (tn_b != NULL) tn_b->parent = tn_3;

  // move tn_3 below tn_4
  tn_4->parent = tn_3->parent;
  if (tn_4->parent != NULL) {
    // stitch up the child ptr from tn_4's new parent
    if (tn_4->parent->left == tn_3) {
      tn_4->parent->left = tn_4;
    } else {
      assert(tn_4->parent->right == tn_3);
      tn_4->parent->right = tn_4;
    }
  }

  tn_3->parent = tn_4;
  tn_4->left = tn_3;

  // reset the height.
  ResetHeightUpwards(tn_3);
}

void RL(AVLTreeNodePtr tn_3) {
  AVLTreeNodePtr tn_5 = tn_3->right, tn_4 = tn_5->left,
    tn_c = tn_4->right;

  // move 4 up to be right-child of tn
  tn_3->right = tn_4;
  tn_4->parent = tn_3;

  // move 5 down to be right-child of 4
  tn_4->right = tn_5;
  tn_5->parent = tn_4;

  // move c to be left-child of 5
  tn_5->left = tn_c;
  if (tn_c != NULL) tn_c->parent = tn_5;

  // reset the heights upwards.
  ResetHeightUpwards(tn_5);
}

void LR(AVLTreeNodePtr tn_5) {
  AVLTreeNodePtr tn_3 = tn_5->left,
    tn_4 = tn_3->right,
    tn_b = tn_4->left;

  // move 4 up to be left-child of tn.
  tn_5->left = tn_4;
  tn_4->parent = tn_5;

  // move 3 down to be left-child of 4.
  tn_4->left = tn_3;
  tn_3->parent = tn_4;

  // move b to be right-child of 3
  tn_3->right = tn_b;
  if (tn_b != NULL) tn_b->parent = tn_3;

  // reset the heights upwards.
  ResetHeightUpwards(tn_3);
}

void LL(AVLTreeNodePtr tn_5) {
  AVLTreeNodePtr tn_4 = tn_5->left,
    tn_c = tn_4->right;

  // move c to be left-child of tn_5.
  tn_5->left = tn_c;
  if (tn_c != NULL) tn_c->parent = tn_5;

  // move tn_5 below tn_4.
  tn_4->parent = tn_5->parent;
  if (tn_4->parent != NULL) {
    // stitch up the child ptr from tn_4's new parent
    if (tn_4->parent->left == tn_5) {
      tn_4->parent->left = tn_4;
    } else {
      assert(tn_4->parent->right == tn_5);
      tn_4->parent->right = tn_4;
    }
  }
  tn_5->parent = tn_4;
  tn_4->right = tn_5;

  // reset the height.
  ResetHeightUpwards(tn_5);
}

void ResetHeightUpwards(AVLTreeNodePtr tn) {
  if (tn == NULL)
    return;

  // If leaf, height is 1.
  if ((tn->left == NULL) && (tn->right == NULL)) {
    tn->height = 1U;
    ResetHeightUpwards(tn->parent);
  }

  // Non-leaf; calculate height based on children.
  uint32_t maxchild = 0;
  if (tn->left != NULL)
    maxchild = tn->left->height;

  if ((tn->right != NULL) && (tn->right->height > maxchild))
    maxchild = tn->right->height;

  tn->height = maxchild + 1U;
  ResetHeightUpwards(tn->parent);
}

int32_t GetBalance(AVLTreeNodePtr tn) {
  int32_t height_left = 0, height_right = 0;

  if (tn->left != NULL)
    height_left = tn->left->height;

  if (tn->right != NULL)
    height_right = tn->right->height;

  return height_left - height_right;
}

int RecursiveLookupInAVLTree(AVLTreeNodePtr tn,
                             void *payload_search,
                             void **payload_return,
                             TreePayloadComparatorFnPtr cfp) {
  if (tn == NULL)
    return 0;

  int retval = cfp(payload_search, tn->payload);
  if (retval == 0) {
    // match
    *payload_return = tn->payload;
    return 1;
  }

  if (retval == -1) {
    // look left
    return RecursiveLookupInAVLTree(tn->left, payload_search,
                                    payload_return, cfp);
  }
  return RecursiveLookupInAVLTree(tn->right, payload_search,
                                  payload_return, cfp);
}

int RecursiveRemoveFromAVLTree(AVLTreeNodePtr tn,
                               void *pl,
                               TreePayloadComparatorFnPtr cfp,
                               TreePayloadFreeFnPtr pffp,
                               AVLTreePtr tree) {
  // Have I bottomed out of the tree?
  if (tn == NULL)
    return 0;

  // Have I found the node? If not, descend.
  switch (cfp(pl, tn->payload)) {
  case -1:
    return RecursiveRemoveFromAVLTree(tn->left, pl, cfp, pffp, tree);
  case +1:
    return RecursiveRemoveFromAVLTree(tn->right, pl, cfp, pffp, tree);
  }

  // We're here! Case 1: the node is a leaf, or has only one child.
  // Remove it, recalculate height starting at parent.
  if ((tn->left == NULL) || (tn->right == NULL)) {
    RemoveLeafOrChildFromAVLTree(tn, pffp, tree);
    return 1;
  }

  // Case 2: the node to delete has two children. Find the inorder
  // successor of the node, within the subtree (don't ascend).
  AVLTreeNodePtr successor = tn->right;
  while (successor->left != NULL)
    successor = successor->left;

  AVLTreeNodePtr successor_parent = successor->parent;

  // Delete successor from the tree (will be the simple case above),
  // but preserve its payload.
  void *payload = successor->payload;
  int res = RecursiveRemoveFromAVLTree(successor,
                                       successor->payload,
                                       cfp,
                                       &NullAVLPayloadDelete,
                                       tree);
  assert(res);

  // Replace self's payload with the saved one.
  pffp(tn->payload);
  tn->payload = payload;

  // If self's new balance is 2, and left subtree is 0, must do
  // a single right rotation via a LL().
  if ((GetBalance(tn) == +2) && (tn->left != NULL) &&
      (GetBalance(tn->left) == 0)) {
    LL(tn);
  }

  // Do a rebalance check on successor's parent.
  if (successor_parent != NULL)
    Rebalance(successor_parent, tree);

  // Success!
  return 1;
}

AVLTreeNodePtr GetSuccessor(AVLTreeNodePtr tn) {
  if (tn == NULL)
    return NULL;

  // if there's a right branch, return its minimum
  if (tn->right != NULL) {
    tn = tn->right;
    while (tn->left != NULL)
      tn = tn->left;
    return tn;
  }

  // we need to go up.  if we came from the right, keep
  // going up.
  AVLTreeNodePtr y = tn->parent;
  while ((y != NULL) && (tn == y->right)) {
    tn = y;
    y = y->parent;
  }
  return y;
}

AVLTreeNodePtr GetPredecessor(AVLTreeNodePtr tn) {
  if (tn == NULL)
    return NULL;

  // if there's a left branch, return its max
  if (tn->left != NULL) {
    tn = tn->left;
    while (tn->right != NULL)
      tn = tn->right;
    return tn;
  }

  // we need to go up.  if we came from the left, keep
  // going up.
  AVLTreeNodePtr y = tn->parent;
  while ((y != NULL) && (tn == y->left)) {
    tn = y;
    y = y->parent;
  }
  return y;
}

int AVLTreeIteratorDoNext(AVLTreeIterPtr iter, int do_advance) {
  // defensive programming
  assert(iter != NULL);

  AVLTreeNodePtr tmp;
  switch (iter->order) {
  case INORDER:
    // inorder goes from successor to successor
    tmp = GetSuccessor(iter->node);
    if (tmp != NULL) {
      if (do_advance == 1)
        iter->node = tmp;
      return 1;
    }
    return 0;
  case INORDER_REVERSE:
    // inorder_reverse goes predecessor to predecessor
    tmp = GetPredecessor(iter->node);
    if (tmp != NULL) {
      if (do_advance == 1)
        iter->node = tmp;
      return 1;
    }
    return 0;
  case PREORDER:
    // Preorder is more complex to do iteratively.  If there
    // is a left, or alternatively right, child, that's next.
    if (iter->node->left != NULL) {
      if (do_advance == 1)
        iter->node = iter->node->left;
      return 1;
    }
    if (iter->node->right != NULL) {
      if (do_advance == 1)
        iter->node = iter->node->right;
      return 1;
    }

    // We have to ascend up, as long as we're coming from right.  Else,
    // once we come up from left, go back down right and that's next.
    AVLTreeNodePtr p = iter->node->parent;
    AVLTreeNodePtr x = iter->node;

    // keep ascending while coming from right, or while there
    // is no right
    while (p != NULL) {
      if ((p->right == NULL) || (p->right == x)) {
        x = p;
        p = p->parent;
        continue;
      }
      // Found it; it's the right node.
      if (do_advance == 1)
        iter->node = p->right;
      return 1;
    }
    // Nope, return fail.
    return 0;
  }
  // should never get here
  assert(0);
  return 0;
}

void RemoveLeafOrChildFromAVLTree(AVLTreeNodePtr tn,
                                  TreePayloadFreeFnPtr pffp,
                                  AVLTreePtr tree) {
  AVLTreeNodePtr promote_me;
  if (tn->left == NULL)
    promote_me = tn->right;
  else
    promote_me = tn->left;

  // Stitch up the parent to point to the promotion.
  if (tn->parent == NULL) {
    // no parent, so need to change the root node of tree to
    // the promotion node.
    tree->root = promote_me;
  } else {
    // is a parent, so update it to the promotion node.
    if (tn->parent->right == tn)
      tn->parent->right = promote_me;
    else
      tn->parent->left = promote_me;
  }

  // Stitch up the promotion to point to the parent.
  if (promote_me != NULL) {
    promote_me->parent = tn->parent;
  }

  // Recalc heights, starting at parent.  Promotion is
  // unperturbed, so no need to recalculate it.
  ResetHeightUpwards(tn->parent);

  // Free my payload.
  pffp(tn->payload);

  // Free me.
  free(tn);

  // Success!
}
