/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _HW1_AVLTREE_H_
#define _HW1_AVLTREE_H_

#include <stdint.h>  // for uint64_t

// An AVL tree is a balanced binary search tree.  We provide the
// interface here; your job is to fill in the implementation holes
// in AVLTree.c.

// To hide the implementation of AVLTree, we declare the
// "struct avl_head" structure here, but we define the structure in
// the internal header AVLTree_priv.h. This lets us define a
// pointer to AVLTree as a new type but leave the implementation
// details of AVLTree opaque to the customer.
struct avltree_head;
typedef struct avltree_head *AVLTreePtr;

// When a customer frees an AVLTree, they need to pass a pointer
// to a function that frees the payload of a tree node.
// We invoke the pointed-to function once for each node in the
// tree, before the AVLTreePtr itself is freed.
typedef void(*TreePayloadFreeFnPtr)(void *payload);

// When comparing two elements of an AVLTree, such as during
// insertion or retrieval, customers must pass in a comparator
// function.  The function accepts two payload pointers as arguments
// and returns an integer that is:
//
//    -1  if payload_a < payload_b
//     0  if payload_a == payload_b
//     1  if payload_a > payload_b
typedef int(*TreePayloadComparatorFnPtr)(void *payload_a, void *payload_b);

// Allocate and return a new AVLTree.  The caller takes
// responsibility for eventually calling FreeAVLTree to
// free any memory associated with the tree.
//
// Arguments: none.
//
// Returns: NULL on error, non-NULL on success.
AVLTreePtr AllocateAVLTree(void);

// Free an AVLTree that was previously allocated by
// AllocateAVLTree.
//
// Arguments:
//
// - tree: the tree  to free.  It is unsafe to use
//   "tree" after this function returns.
//
// - payload_free_function: a pointer to a payload freeing
//   function; see above for details.
void FreeAVLTree(AVLTreePtr tree,
                 TreePayloadFreeFnPtr payload_free_function);

// Return the number of elements in the tree.
//
// Arguments:
//
// - tree:  the tree to query
//
// Returns:
//
// - tree size
uint64_t NumElementsInAVLTree(AVLTreePtr tree);

// Inserts a new element to the tree.
//
// Arguments:
//
// - tree: the AVLTree to insert into
//
// - payload: the payload to push; it's up to the caller to
//   interpret and manage the memory of the payload.
//
// - comparator_function: a comparator function used
//   by the tree implementation to compare nodes when finding
//   the right place to create the new node.  This is needed
//   since an AVL tree is a binary tree (and hence is sorted).
//
// Returns 0 on failure, +1 on success.
int InsertIntoAVLTree(AVLTreePtr tree,
                      void *payload,
                      TreePayloadComparatorFnPtr comparator_function);

// Look up an element in the tree.
//
// Arguments:
//
// - tree: the AVL tree to lookup in.
//
// - payload: a payload pointer to do the search on, using the
//   comparator_function.
//
// - payload_return: a return parameter through which the matching
//   payload pointer is returned.
//
// - comparator_function: a comparator function used to compare
//   nodes to payload_search.
//
// Returns:
//
// - if a match is found, returns 1, and returns the payload through
//   payload_return
//
// - if no match is found, returns 0
int LookupInAVLTree(AVLTreePtr tree,
                    void *payload_search,
                    void **payload_return,
                    TreePayloadComparatorFnPtr comparator_function);

// Removes an element from the AVLTree.
//
// Arguments:
//
// - tree: the AVLTree to remove from.
//
// - remove_me: a pointer to the payload to remove from the tree.
//
// - comparator_function: a comparator function used by the tree
//   implementation to find the right node to return.
//
// - payload_free_function: a pointer to a payload freeing function;
//   the tree implementation frees the payload it removes, assuming
//   it successfully finds it.
//
// Returns 0 on failure (the remove_me element could not be found),
// +1 on success (the element was found, removed, and freed).
int RemoveFromAVLTree(AVLTreePtr tree,
                      void *remove_me,
                      TreePayloadComparatorFnPtr comparator_function,
                      TreePayloadFreeFnPtr payload_free_function);


// An AVL tree supports the notion of an iterator, similar to a Java
// iterator.  A tree supports four kinds of iteration:
//
// - inorder traversal: this is the most common; it traverses
//     the tree nodes in ascending order by recursively
//     inorder traversing the left subtree followed by
//     visiting the root followed by recursively inorder
//     traversing the right subtree.
//
// - inorder reverse: same as inorder, but traverses in
//     descending order instead of ascending order.
//
// - preorder: visit the node, then left tree, then right tree.
//     useful for generating polish notation (prefix) expression trees,
//     or for copying a tree with minimal insertions in the copy
//
// - postorder: visit left tree, then right tree, then node.
typedef enum {
  INORDER = 0, INORDER_REVERSE, PREORDER
} AVLTreeIteratorOrder;

// To use an iterator, firse use AVLTreeMakeIterator to
// manufacture one, then repeatedly use AVLTreeIteratorHasNext() or
// AVLTreeIteratorGetNext() to iterate through it.  Note that you
// should not modify a tree while using an iterator.  Once
// you have finished with an iterator, call AVLTreeIteratorFree
// to delete it.
struct avltree_iter;
typedef struct avltree_iter *AVLTreeIterPtr;  // hides implementation.

// Manufacture an iterator for the tree.  Customers can use the
// iterator to advance forwards in the specified order.  Caller is
// responsible for eventually calling AVLTreeIteratorFree to free
// memory associated with the iterator.
//
// Arguments:
//
// - tree: the tree from which we'll return an iterator
//
// - order: which traversal order to use
//
// Returns NULL on failure (out of memory, tree is empty) or
// non-NULL on success.
AVLTreeIterPtr AVLTreeMakeIterator(AVLTreePtr tree,
                                   AVLTreeIteratorOrder order);

// When you're done with an iterator, you need to free it by
// calling this function.
//
// Arguments:
//
// - iter: the iterator to free. Don't use it after freeing it.
void AVLTreeIteratorFree(AVLTreeIterPtr iter);

// Test whether the iterator can advance forward.
//
// Arguments:
//
// - iter: the iterator
//
// Returns:
//
// - +1 if the iterator can advance (is not at end of the traversal)
//
// - 0 if the iterator cannot advance (is at the end of the traversal)
int AVLTreeIteratorHasNext(AVLTreeIterPtr iter);

// Advance the iterator.
//
// Arguments:
//
// - iter: the iterator
//
// Returns:
//
// +1: if the advance worked
//
// 0: otherwise (e.g., at the end).
int AVLTreeIteratorNext(AVLTreeIterPtr iter);

// Returns the payload of the tree node that the iterator currently
// points at.
//
// Arguments:
//
// - iter: the iterator to fetch the payload from.
//
// - payload: a "return parameter" through which the payload
//   is returned.
void AVLTreeIteratorGetPayload(AVLTreeIterPtr iter, void **payload);

#endif  // _HW1_AVLTREE_H_
