/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE 333 course project sequence
 *  (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <stdint.h>

#include "AVLTree.h"

// This is the type of payload we store in our AVLtree: a
// key/value pair.  We'll write a PayloadComparator function
// to compare against keys.  This way, we can insert by key and
// lookup by key.  We can also update the value within the
// record freely, since it doesn't update the position of the
// element within the tree.  Here, we'll make the value be
// a dynamically allocated string.
typedef struct {
  int64_t       key;     // keys are 64 bit integers
  unsigned char *value;  // values are strings (or NULL)
} KeyValPayload, *KeyValPayloadPtr;

// prototype of our payload free function
void KeyValPayloadFree(void *payload);

// prototype of our payload comparator function
int KeyValPayloadComparator(void *p1, void *p2);

// a helper function to alloc/insert a record
void AllocAndInsert(AVLTreePtr tree, int i);

// our main function; here, we demonstrate how to use some
// of the AVLTree functions.
int main(int argc, char **argv) {
  // Create an AVL tree.
  AVLTreePtr tree = AllocateAVLTree();
  assert(tree != NULL);
  assert(NumElementsInAVLTree(tree) == 0U);

  // Insert some elements into the tree; we'll insert
  // a bunch of even keys, then a bunch of odd keys.
  for (int i=0; i<1000; i+=2) {
    AllocAndInsert(tree, i);
  }
  assert(NumElementsInAVLTree(tree) == 500U);

  // Insert the odd elements into the tree.
  for (int i=1; i<1000; i+=2) {
    AllocAndInsert(tree, i);
  }
  assert(NumElementsInAVLTree(tree) == 1000U);

  // Walk the tree, making sure everything is in sorted order.
  AVLTreeIterPtr atip = AVLTreeMakeIterator(tree, INORDER);
  int64_t expect_next = 0U;
  while (1) {
    // Get the next payload
    KeyValPayloadPtr kvpp = NULL;
    AVLTreeIteratorGetPayload(atip, (void **) &kvpp);
    assert(kvpp != NULL);

    // Make sure it's what we expect.
    char buf[10];
    assert(kvpp->key == expect_next);
    sprintf(buf, "%d", (int) expect_next);
    assert(strcmp(buf, (char *) kvpp->value) == 0);

    // Advance to the next.
    expect_next++;
    if (expect_next == 1000U) {
      assert(AVLTreeIteratorHasNext(atip) == 0);
      break;
    } else {
      assert(AVLTreeIteratorHasNext(atip) == 1);
    }
    assert(AVLTreeIteratorNext(atip) == 1);
  }

  // Clean up the iterator.
  AVLTreeIteratorFree(atip);

  // Manually update an item, just to show how to do it.
  KeyValPayload kvp = { 10, NULL };  // what to search for
  KeyValPayloadPtr kvpp = NULL;      // pointer to found record
  assert(LookupInAVLTree(tree,
                         (void *) &kvp,
                         (void **) &kvpp,
                         &KeyValPayloadComparator) == 1);
  sprintf((char *) kvpp->value, "Updated!");  // updates in place

  // Delete a value.
  assert(RemoveFromAVLTree(tree,
                           (void *) &kvp,
                           &KeyValPayloadComparator,
                           &KeyValPayloadFree) == 1);

  // Make sure it's deleted.
  assert(LookupInAVLTree(tree,
                         (void *) &kvp,
                         (void **) &kvpp,
                         &KeyValPayloadComparator) == 0);
  assert(RemoveFromAVLTree(tree,
                           (void *) &kvp,
                           &KeyValPayloadComparator,
                           &KeyValPayloadFree) == 0);
  
  // Clean up the tree.
  FreeAVLTree(tree, &KeyValPayloadFree);
  return EXIT_SUCCESS;
}

void KeyValPayloadFree(void *payload) {
  KeyValPayloadPtr kvpp = (KeyValPayloadPtr) payload;
  if (kvpp != NULL) {
    if (kvpp->value != NULL) {
      free(kvpp->value);
      kvpp->value = NULL;
    }
    free(kvpp);
  }
}

// prototype of our payload comparator function
int KeyValPayloadComparator(void *p1, void *p2) {
  assert(p1 != NULL);
  assert(p2 != NULL);

  int64_t i1, i2;
  i1 = ((KeyValPayloadPtr) p1)->key;
  i2 = ((KeyValPayloadPtr) p2)->key;

  if (i1 > i2)
    return 1;
  if (i1 < i2)
    return -1;
  return 0;
}

void AllocAndInsert(AVLTreePtr tree, int i) {
  // Allocate the payload, the string within it, then initialize them.
  KeyValPayloadPtr kvpp = (KeyValPayloadPtr) malloc(sizeof(KeyValPayload));
  assert(kvpp != NULL);
  kvpp->value = (unsigned char *) malloc(10*sizeof(char));
  assert(kvpp->value != NULL);
  kvpp->key = (int64_t) i;
  sprintf((char *) kvpp->value, "%d", i);

  // Insert the payload into the tree.
  assert(InsertIntoAVLTree(tree,
                           (void *) kvpp,
                           &KeyValPayloadComparator) == 1);
}
