/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <unistd.h>
#include <errno.h>
#include <sys/select.h>

extern "C" {
  #include "./AVLTree.h"
  #include "./AVLTree_priv.h"
}

#include "./test_suite.h"
#include "./test_avltree.h"

namespace hw1 {

static unsigned int free_count = 0U;
void AVLPayloadFreeFunction(void *payload) {
  // Do nothing but verify the payload is non-NULL and
  // increment the free count.
  ASSERT_NE(payload, (void *) NULL);
  free_count++;
}

int AVLPayloadComparator(void *a, void *b) {
  uint64_t i1 = (uint64_t) a;
  uint64_t i2 = (uint64_t) b;

  if (i1 > i2)
    return 1;
  if (i1 < i2)
    return -1;
  return 0;
}


TEST_F(Test_AVLTree, TestAVLTreeBasic) {
  // Try creating a tree.
  AVLTreePtr t = AllocateAVLTree();
  ASSERT_NE((AVLTreePtr) NULL, t);
  ASSERT_EQ(0U, NumElementsInAVLTree(t));
  ASSERT_EQ(NULL, t->root);

  // Add (out of order) a few things to the tree.
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 10U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 20U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 30U, &AVLPayloadComparator));

  // Check the tree for correctness.
  ASSERT_EQ(2U, t->root->height);
  ASSERT_EQ(20U, (uint64_t) t->root->payload);
  ASSERT_NE((AVLTreeNodePtr) NULL, t->root->left);
  ASSERT_NE((AVLTreeNodePtr) NULL, t->root->right);
  ASSERT_EQ(t->root, t->root->left->parent);
  ASSERT_EQ(t->root, t->root->right->parent);
  ASSERT_EQ(10U, (uint64_t) t->root->left->payload);
  ASSERT_EQ(30U, (uint64_t) t->root->right->payload);
  ASSERT_EQ(1U, t->root->left->height);
  ASSERT_EQ(1U, t->root->right->height);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->left->left);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->left->right);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->right->left);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->right->right);

  // Insert some more.
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 15U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 17U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 19U, &AVLPayloadComparator));

  // Now that we've done a bunch of rotations, test in more depth.
  ASSERT_EQ(17U, (uint64_t) t->root->payload);
  ASSERT_EQ(20U, (uint64_t) t->root->right->payload);
  ASSERT_EQ(30U, (uint64_t) t->root->right->right->payload);
  ASSERT_EQ(19U, (uint64_t) t->root->right->left->payload);
  ASSERT_EQ(15U, (uint64_t) t->root->left->payload);
  ASSERT_EQ(10U, (uint64_t) t->root->left->left->payload);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->parent);
  ASSERT_EQ((AVLTreeNodePtr) NULL, t->root->left->right);
  ASSERT_EQ(t->root, t->root->left->parent);
  ASSERT_EQ(t->root, t->root->right->parent);
  ASSERT_EQ(t->root->right, t->root->right->right->parent);
  ASSERT_EQ(t->root->right, t->root->right->left->parent);
  ASSERT_EQ(t->root, t->root->left->parent);
  ASSERT_EQ(t->root->left, t->root->left->left->parent);
  ASSERT_EQ(3U, t->root->height);
  ASSERT_EQ(2U, t->root->right->height);
  ASSERT_EQ(1U, t->root->right->right->height);
  ASSERT_EQ(1U, t->root->right->left->height);
  ASSERT_EQ(2U, t->root->left->height);
  ASSERT_EQ(1U, t->root->left->left->height);

  // Exercise the last rotation case.
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 40U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 35U, &AVLPayloadComparator));
  ASSERT_EQ(35U, (uint64_t) t->root->right->right->payload);
  ASSERT_EQ(30U, (uint64_t) t->root->right->right->left->payload);
  ASSERT_EQ(40U, (uint64_t) t->root->right->right->right->payload);
  ASSERT_EQ(4U, t->root->height);
  ASSERT_EQ(t->root->right->right,
            t->root->right->right->right->parent);
  ASSERT_EQ(t->root->right->right,
            t->root->right->right->left->parent);

  // Try deleting the tree.
  free_count = 0U;
  FreeAVLTree(t, &AVLPayloadFreeFunction);
  ASSERT_EQ(8U, free_count);
  t = NULL;
}

TEST_F(Test_AVLTree, TestAVLTreeAddDelete) {
  // Create a tree.
  AVLTreePtr t = AllocateAVLTree();
  ASSERT_NE((AVLTreePtr) NULL, t);
  ASSERT_EQ(0U, NumElementsInAVLTree(t));
  ASSERT_EQ(NULL, t->root);

  // Add to it.
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 10U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 20U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 30U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 15U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 17U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 19U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 40U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 35U, &AVLPayloadComparator));

  // Delete some elements from the tree.
  free_count = 0U;
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 17U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 17U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1U, free_count);
  ASSERT_EQ(19U, (uint64_t) t->root->payload);
  ASSERT_EQ(35U, (uint64_t) t->root->right->payload);
  ASSERT_EQ(20U, (uint64_t) t->root->right->left->payload);
  ASSERT_EQ(30U, (uint64_t) t->root->right->left->right->payload);
  ASSERT_EQ(40U, (uint64_t) t->root->right->right->payload);

  // Delete the rest, one by one.
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 20U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 20U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 40U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 40U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 15U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 15U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 35U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 35U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 19U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 19U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 30U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 30U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) 10U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(0, RemoveFromAVLTree(t, (void *) 10U, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(8U, free_count);

  // Delete the tree.
  free_count = 0U;
  FreeAVLTree(t, &AVLPayloadFreeFunction);
  ASSERT_EQ(0U, free_count);
  t = NULL;
}

TEST_F(Test_AVLTree, TestAVLTreeTraversals) {
  // Create a tree.
  AVLTreePtr t = AllocateAVLTree();
  ASSERT_NE((AVLTreePtr) NULL, t);
  ASSERT_EQ(0U, NumElementsInAVLTree(t));
  ASSERT_EQ(NULL, t->root);

  // Add to it.
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 10U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 20U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 30U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 15U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 17U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 19U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 40U, &AVLPayloadComparator));
  ASSERT_EQ(1, InsertIntoAVLTree(t, (void *) 35U, &AVLPayloadComparator));

  // Try inorder traversal.
  AVLTreeIterPtr it = AVLTreeMakeIterator(t, INORDER);
  ASSERT_NE((AVLTreeIterPtr) NULL, it);
  void *payload = NULL;
  int itercount = 0;
  uint64_t last = 0;
  while (1) {
    AVLTreeIteratorGetPayload(it, &payload);
    ASSERT_LT(last, (uint64_t) payload);
    last = (uint64_t) payload;
    itercount++;
    if (!AVLTreeIteratorNext(it))
      break;
  }
  ASSERT_EQ(8, itercount);
  // Free the iterator.
  AVLTreeIteratorFree(it);

  // Try inorder-reverse traversal.
  it = AVLTreeMakeIterator(t, INORDER_REVERSE);
  ASSERT_NE((AVLTreeIterPtr) NULL, it);
  itercount = 0;
  last = 99999;
  while (1) {
    AVLTreeIteratorGetPayload(it, &payload);
    ASSERT_GT(last, (uint64_t) payload);
    last = (uint64_t) payload;
    itercount++;
    if (!AVLTreeIteratorNext(it))
      break;
  }
  ASSERT_EQ(8, itercount);
  // Free the iterator.
  AVLTreeIteratorFree(it);

  // Try preorder traversal.
  it = AVLTreeMakeIterator(t, PREORDER);
  ASSERT_NE((AVLTreeIterPtr) NULL, it);
  // Step through, make sure each step is right.
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(17U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(15U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(10U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(20U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(19U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(35U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(30U, (uint64_t) payload);
  ASSERT_EQ(1, AVLTreeIteratorNext(it));
  AVLTreeIteratorGetPayload(it, &payload);
  ASSERT_EQ(40U, (uint64_t) payload);
  ASSERT_EQ(0, AVLTreeIteratorNext(it));
  AVLTreeIteratorFree(it);

  // Delete the tree.
  free_count = 0U;
  FreeAVLTree(t, &AVLPayloadFreeFunction);
  ASSERT_EQ(8U, free_count);
  t = NULL;
}

TEST_F(Test_AVLTree, TestAVLTreeBigRandom) {
  // Create a tree.
  AVLTreePtr t = AllocateAVLTree();
  ASSERT_NE((AVLTreePtr) NULL, t);
  ASSERT_EQ(0U, NumElementsInAVLTree(t));
  ASSERT_EQ(NULL, t->root);

  // Add to it, keep track of three for hand-deletion.
  uint64_t added_count = 0;
  uint64_t a, b, c;
  for (int i = 0; i < 100000; i++) {
    uint64_t next = rand();  // NOLINT(runtime/threadsafe_fn)
    if (i == 0) a = next;
    if (i == 1) b = next;
    if (i == 2) c = next;
    void *retptr;
    if (LookupInAVLTree(t, (void *) next, &retptr,
                        &AVLPayloadComparator)) {
      ASSERT_EQ(next, (uint64_t) retptr);
    } else {
      ASSERT_EQ(1, InsertIntoAVLTree(t,
                                     (void *) (uint64_t) next,
                                     &AVLPayloadComparator));
      added_count++;
    }
  }

  // Delete three.
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) a, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) b, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  ASSERT_EQ(1, RemoveFromAVLTree(t, (void *) c, &AVLPayloadComparator,
                                 &AVLPayloadFreeFunction));
  added_count -= 3;

  // Try inorder traversal.
  AVLTreeIterPtr it = AVLTreeMakeIterator(t, INORDER);
  ASSERT_NE((AVLTreeIterPtr) NULL, it);
  void *payload = NULL;
  uint64_t itercount = 0;
  uint64_t last = 0;
  while (1) {
    // verify the tree is in balance throughout
    int balance = GetBalance(it->node);
    ASSERT_LT(-2, balance);
    ASSERT_GT(+2, balance);

    // verify the payload is monotonically increasing
    AVLTreeIteratorGetPayload(it, &payload);
    ASSERT_LT(last, (uint64_t) payload);
    last = (uint64_t) payload;
    itercount++;
    if (!AVLTreeIteratorNext(it))
      break;
  }
  // make sure we found everything we inserted
  ASSERT_EQ(added_count, itercount);

  // Free the iterator.
  AVLTreeIteratorFree(it);

  // Free the tree.
  free_count = 0U;
  FreeAVLTree(t, &AVLPayloadFreeFunction);
  ASSERT_EQ(added_count, free_count);
  t = NULL;
}

}  // namespace hw1
